#include "ARP2GTSP.h"
#include "Heap.h"

static Arc *CreateArc(int To, int Cost, int Required, Arc * Next);
static void Dijkstra(Vertex * Source);
static void WritePath(FILE *out, Vertex * v);

static Vertex *V;
static int VSize;

int main(int argc, char *argv[])
{
    int RequiredVertices = 0;
    int RequiredEdges = 0;
    int NotRequiredEdges = 0;
    int RequiredArcs = 0;
    int Edges = 0;
    int WSize;
    int Clusters;
    int Dimension = 0;
    int *GTour;
    int i, j, k, x, y;
    Wertex *W;
    Vertex *v, *v1, *v2;
    Vertex *From, *To, *LastTo;
    Arc *a;
    int **C;
    char *Class, Line[256];
    FILE *in, *out;
    
    if (argc != 5) {
        printf("Usage: ExtractRoute Class Id gtour-file route-file\n");;
        exit(0);
    }
    Class = argv[1];
    sprintf(Line, "%s/%s", Class, argv[2] + (Class[0] == 'G' ? 4 : 0));
    if (!(in = fopen(Line, "r"))) {
        printf("No such file: %s\n", Line);
        exit(0);
    }
    while (fgets(Line, 256, in))
        if (!strncmp(Line, "VERTICES", strlen("VERTICES")))
            break;
    sscanf(Line, "VERTICES :%d\n", &VSize);
    while (fgets(Line, 256, in))
        if (!strncmp(Line, "ARISTAS_REQ", strlen("ARISTAS_REQ")))
            break;
    sscanf(Line, "ARISTAS_REQ :%d\n", &RequiredEdges);
    while (fgets(Line, 256, in))
        if (!strncmp(Line, "ARISTAS_NOREQ", strlen("ARISTAS_NOREQ")))
            break;
    sscanf(Line, "ARISTAS_NOREQ :%d\n", &NotRequiredEdges);
    while (fgets(Line, 256, in))
        if (!strncmp(Line, "LISTA_A", strlen("LISTA_A")))
            break;
    
    assert(V = (Vertex *) calloc(1 + VSize, sizeof(Vertex)));
    for (i = 1; i <= VSize; i++) {
        v = &V[i];
        v->Id = i;
        v->InDegree = v->OutDegree = 0;
    }
    for (i = 0; i < RequiredEdges; i++) {
        int a, b, c, d = -1;
        
        fscanf(in, "(%d,%d)   coste  %d %d\n", &a, &b, &c, &d);
        v1 = &V[a];
        v2 = &V[b];
        if (c >= 0 && c != 9999 && c != 99999999) {
            v1->FirstArc = CreateArc(b, c, 1, v1->FirstArc);
            RequiredArcs++;
            v1->OutDegree++;
            v2->InDegree++;
        }
        if (d >= 0 && d != 9999 && d != 99999999) {
            v2->FirstArc = CreateArc(a, d, 1, v2->FirstArc);
            RequiredArcs++;
            v2->OutDegree++;
            v1->InDegree++;
        }
        if (c == d)
            Edges++;
    }
    fgets(Line, 256, in);
    for (i = 0; i < NotRequiredEdges; i++) {
        int a, b, c, d = -1;
        fscanf(in, "(%d,%d)   coste  %d %d\n", &a, &b, &c, &d);
        v1 = &V[a];
        v2 = &V[b];
        if (c >= 0 && c != 9999 && c != 99999999)
            v1->FirstArc = CreateArc(b, c, 0, v1->FirstArc);
        if (d >= 0 && d != 9999 && d != 99999999)
            v2->FirstArc = CreateArc(a, d, 0, v2->FirstArc);
    }
    fgets(Line, 256, in);
    for (i = 1; i <= VSize; i++) {
        v = &V[i];
        if (v->InDegree == 0 && v->OutDegree == 0) {
            v->Required = 1;
            RequiredVertices++;
        }
    }
    fclose(in);
    
    WSize = RequiredVertices + RequiredArcs;
    if (Class[0] == 'W')
        Clusters = WSize - RequiredEdges;
    else
        Clusters = WSize - Edges;
    
    
    sprintf(Line, "GTSP/%s.gtsp", argv[2]);
    assert(out = fopen(Line, "w"));
    assert(W = (Wertex *) calloc(1 + WSize, sizeof(Wertex)));
    for (i = 1; i <= WSize; i++)
        W[i].Id = i;
    
    k = WSize;
    WSize = 0;
    for (i = 1; i <= VSize; i++) {
        v = &V[i];
        if (v->Required) {
            WSize++;
            W[WSize].From = W[WSize].To = v;
            W[WSize].Cost = 0;
        } else {
            for (a = v->FirstArc; a; a = a->Next) {
                if (a->Required) {
                    WSize++;
                    W[WSize].From = v;
                    W[WSize].To = &V[a->To];
                    W[WSize].Cost = a->Cost;
                }
            }
        }
    }
    if (WSize != k) {
        printf("ERROR: WSize: %d != OldWSize: %d\n", WSize, k);
        exit(1);
    }
    
    assert(C = (int **) malloc((WSize + 1) * sizeof(int *)));
    for (i = 1; i <= WSize; i++) {
        assert(C[i] = (int *) malloc((WSize + 1) * sizeof(int)));
        for (j = 1; j <= WSize; j++)
            C[i][j] = 999999;
    }
    
    MakeHeap(VSize);
    for (x = 1; x <= WSize; x++) {
        Dijkstra(W[x].To);
        for (y = 1; y <= WSize; y++)
            if (y != x)
                C[x][y] = W[y].From->Cost + W[y].Cost;
    }
    
    if (!(in = fopen(argv[3], "r"))) {
        printf("Cannot open gtour file: %s\n", argv[3]);
        exit(0);
    }
    while (fgets(Line, 256, in))
        if (!strncmp(Line, "DIMENSION", strlen("DIMENSION")))
            break;
    sscanf(Line, "DIMENSION : %d\n", &Dimension);
    if (Dimension != Clusters) {
        printf("Dimension: %d != Clusters: %d\n", Dimension, Clusters);
        exit(0);
    }
    while (fgets(Line, 256, in))
        if (!strncmp(Line, "TOUR_SECTION", strlen("TOUR_SECTION")))
            break;
    assert(GTour = (int *) malloc((Dimension + 1) * sizeof(int)));
    for (i = 1; i <= Dimension; i++)
        fscanf(in, "%d", &GTour[i]);
    GTour[0] = GTour[Dimension];
    fclose(in);
    
    if (!(out = fopen(argv[4], "w"))) {
        printf("Cannot open tour file %s\n", argv[4]);
        exit(0);
    }
    fprintf(out, "NAME : %s\n", argv[4]);
    fprintf(out, "TYPE : ROUTE\n");
    fprintf(out, "TOUR_SECTION\n");
    LastTo = W[GTour[Dimension]].To;
    MakeHeap(VSize);
    int Cost = 0;
    for (i = 1; i <= Dimension; i++) {
        From = W[GTour[i]].From;
        To = W[GTour[i]].To;
        if (From != LastTo) {
            Dijkstra(LastTo);
            Cost += From->Cost;
            WritePath(out, From->Dad);
        } 
        if (From != To) {
            Cost += W[GTour[i]].Cost;
            fprintf(out, "%d\n", From->Id);
        }
        LastTo = To;
    }
    fprintf(out, "-1\nEOF\n");
    printf("%s: Cost = %d\n", argv[4], Cost);
}

static Arc *CreateArc(int To, int Cost, int Required, Arc * Next)
{
    Arc *a;
    assert(a = (Arc *) malloc(sizeof(Arc)));
    a->To = To;
    a->Cost = Cost;
    a->Required = Required;
    a->Next = Next;
    return a;
}

static void Dijkstra(Vertex * Source)
{
    Vertex *Blue = Source, *v;
    Arc *a;
    int i;

    Blue->Loc = 0;
    Blue->Cost = 0;
    Blue->Dad = 0;
    for (i = 1; i <= VSize; i++) {
        v = &V[i];
        if (v != Blue) {
            v->Dad = Blue;
            v->Cost = 999999;
            HeapLazyInsert(v);
        }
    }
    for (a = Blue->FirstArc; a; a = a->Next) {
        v = &V[a->To];
        v->Dad = Blue;
        v->Cost = a->Cost;
        HeapSiftUp(v);
    }
    while ((Blue = HeapDeleteMin())) {
        for (a = Blue->FirstArc; a; a = a->Next) {
            v = &V[a->To];
            if (!v->Loc || Blue->Cost + a->Cost >= v->Cost)
                continue;
            v->Dad = Blue;
            v->Cost = Blue->Cost + a->Cost;
            HeapSiftUp(v);
        }
    }
}

static Vertex *Last;

static void WritePath(FILE *out, Vertex * v)
{
    if (v) {
        WritePath(out, v->Dad);
        if (v != Last) {
            fprintf(out, "%d\n", v->Id);
            Last = v;
        }
    }
}
